// Derived from code posted here http://www.cs.princeton.edu/~traer/randomarboretum
// Many thanks to the developers of the traer.physics library


import moviemaker.*;
import org.json.*;
import traer.physics.*;
import traer.animation.*;

final float NODE_SIZE = 10;
final float EDGE_LENGTH = 20;
final float EDGE_STRENGTH = 0.2;
final float SPACER_STRENGTH = 1000;

ParticleSystem physics;
Smoother3D centroid;

MovieMaker mm;
boolean movieOn;

Hashtable drawScripts; // JSONArrays of draw commands
Object dateNames[]; // keys for drawScripts
int currentDate; // an index into dateNames
int currentFrame; // an index into current day's json commands

JSONArray drawCode; //JSONArray of draw commands
int drawPtr;
int fadeCounter, fadePeriod;


JSONArray drawScript;
int scriptPtr;
String dateName;
String yesterdayName;
ArrayList branchNodes;
ArrayList branchAges;
ArrayList branchTags;
int maxAge;

color bgColor, altBgColor;
color textColor, altTextColor;

int dateIdx;
String theDate;
String dateOrder[];
PFont dateFont;

int newNodeTimer;
PFont specimenFont,statFont;



// PROCESSING /////////////////////////////////////

void setup()
{
  size( 700, 700 );
  
  movieOn = false;
  
  smooth();
  frameRate( 28 );
  strokeWeight( 2 );
  ellipseMode( CENTER );       
  
  physics = new ParticleSystem( 0, 0.25 );
  centroid = new Smoother3D( 0.97 );
    
  initialize();
  if (movieOn){
      mm = new MovieMaker(this,width,height,"Specimens.mov", MovieMaker.JPEG, MovieMaker.HIGH,30);
  }else{
      print("Not recording today");
  }

  specimenFont = loadFont("ITCFranklinGothicStd-Demi-16.vlw");
  statFont = loadFont("ITCFranklinGothicStd-MedIt-14.vlw");

  //dateFont = loadFont("CstBerlinWest-Original-13.vlw");
  //dateFont = loadFont("CstBerlinWest-Original-70.vlw");
  //dateFont = loadFont("Caspari-70.vlw");
  dateFont = loadFont("AmericanTypewriter-24.vlw");


  bgColor = color(#F5F1DC);
  //altBgColor = color(#424230);
  altBgColor = color(#A7B591);
  textColor = color(#721600);
  altTextColor = color(255,255,255,190);
  fadePeriod = 50;
  
  // Load in the drawing instructions
  String drawData[] = loadStrings("each.js");
  String drawText = drawData[0];
  try {
      drawCode = new JSONArray(drawText);
      branchNodes = new ArrayList();
      branchAges = new ArrayList();
      branchTags = new ArrayList();
      fadeCounter = 0;
      yesterdayName = " ";
  }catch (JSONException e) {println (e.toString());}
}


void draw()
{
  physics.tick( 1.6 ); 
  if ( physics.numberOfParticles() > 1 )
    updateCentroid();
  centroid.tick();
  translate( width/2 , height/2 );
  scale( centroid.z() );
  translate( -centroid.x(), -centroid.y() );
 
  //drawBg();
  background(bgColor);
  drawNetwork();  
  drawOverlay();
  
  /*
  String frameName;
  if (fadeCounter>0){
      frameName = "img/"+yesterdayName+"-####.png";
  }else{
      frameName = "img/"+theDate+"-####.png";
  }
  */

  // Add a new node after every few redraws  
  if (newNodeTimer%3==0){
   nextNode();
  }
  newNodeTimer++;

  //saveFrame(frameName);
  if (movieOn){
      loadPixels();
      mm.addFrame(pixels,width,height);
  }
}

void drawOverlay(){
    if (theDate == null){
        return;
    }

    pushMatrix();
    resetMatrix();  

    // Draw the white box
    fill(255,255,255);
    rect(width-310,height-80,298,68);

    textAlign(RIGHT);
    textFont(specimenFont,16);

    // draw the red date digit
    fill(#971B1E);
    float xMargin = width-20;
    float baseline = height-59;
    if (theDate != null){
        text(theDate, xMargin, baseline);
    }

    // Draw the grey SPECIMEN text
    fill(#5C5D60);
    text("SPECIMEN #",xMargin - textWidth(theDate),baseline);

    baseline+=16;

    // Draw the node count
    textFont(statFont,14);
    text(physics.numberOfParticles()+" nodes",xMargin,baseline);
    popMatrix();
}

void drawBg(){
    pushMatrix();
    resetMatrix();  
    textAlign(CENTER);
    textFont(dateFont,70);


    color textInvis,altTextInvis;
    textInvis = color(red(textColor),green(textColor),blue(textColor),0);
    altTextInvis = color(red(altTextColor),green(altTextColor),blue(altTextColor),0);

    if (fadeCounter>=0){
        if (fadeCounter==fadePeriod){
            color tempColor = bgColor;
            bgColor = altBgColor;
            altBgColor = tempColor;
            
            tempColor = textColor;
            textColor = altTextColor;
            altTextColor = tempColor;            
            textInvis = color(red(textColor),green(textColor),blue(textColor),0);
            altTextInvis = color(red(altTextColor),green(altTextColor),blue(altTextColor),0);
        }
        float fadeProportion = 1.0 - fadeCounter/(float)fadePeriod;

        // Fade the background
        background(interpColor(bgColor,altBgColor,fadeProportion));

        // fade out yesterday date
        fill(interpColor(altTextInvis,altTextColor,fadeProportion));
        if (alpha(interpColor(altTextInvis,altTextColor,fadeProportion))>0){
            text(yesterdayName, width/2, 70);
            //text(fadeCounter+"\n"+fadeProportion+"\n"+red(interpColor(altTextInvis,altTextColor,fadeProportion))+"/"+alpha(interpColor(altTextInvis,altTextColor,fadeProportion)), width/2, 70);
        }

        // overdraw today date
        fill(interpColor(textColor,textInvis,fadeProportion));
        if (alpha(interpColor(textColor,textInvis,fadeProportion))>0){
            text(prettyDate(theDate), width/2, 70);
        }
        
        fadeCounter--;        
    }else{
        background( bgColor );        
        // draw date
        fill(textColor);
        text(prettyDate(theDate), width/2, 70);
    }
 

    popMatrix();
}


void drawNetwork()
{      
  color rootColor = color(150,20,20);
  color newColor = color(140,167,46);
  color oldColor = color(244,204,81);
  color oldAnchorColor = color(#CDD4D8);
  color anchorColor = color(#99b4be);
  
  int numParticles = physics.numberOfParticles();
  int winSize = (int) max(130,numParticles);
  int winPadding = winSize - numParticles;
  if (winPadding<0){winPadding=0;}
  float relativeAge=0.0;

  // draw edges 
  stroke( #696B56 );
  strokeWeight(3);
  beginShape( LINES );
  for ( int i = 0; i < physics.numberOfSprings(); ++i )
  {
    Spring e = physics.getSpring( i );
    Particle a = e.getOneEnd();
    Particle b = e.getTheOtherEnd();
    vertex( a.position().x(), a.position().y() );
    vertex( b.position().x(), b.position().y() );
  }
  endShape();

  // draw vertices
  noStroke();
  for ( int i = 0; i < numParticles; ++i )
  {
    Particle v = physics.getParticle(i);
    color agedColor, invisColor;
    int oldestPossible = 130;//max(130,maxAge);
    float agedness;
    Integer nodeAge;
    String tagName = (String)branchTags.get(i);
    
    //println("Drawing "+tagName);
    if (tagName.equals("body")){ // (tagName.equals("root")){
        // This is the root, color it red
        fill(rootColor); 
        ellipse( v.position().x(), v.position().y(), 2*NODE_SIZE, 2*NODE_SIZE );
        v.setVelocity(8.0,-2.0,0);
    }else{
        // Otherwise, size and color it proportional to age
        nodeAge = (Integer)branchAges.get(i);
        agedness = nodeAge.floatValue()/(float)oldestPossible;
        if (agedness>1.0){agedness=1.0;}
        
        if (tagName.equals("table") || tagName.equals("tr") || tagName.equals("td")){
            agedColor = interpColor(oldAnchorColor,anchorColor,agedness);
        }else{
            agedColor = interpColor(oldColor,newColor,agedness);
        }
        fill(agedColor);        
        ellipse( v.position().x(), v.position().y(), NODE_SIZE+(agedness)*5*NODE_SIZE, NODE_SIZE+(agedness)*5*NODE_SIZE );

        // color in the center once they've arbirarily `bloomed'
        if (agedness>=0.85) {
            fill(255,255,255,130*(agedness-0.85)/0.15);
            ellipse( v.position().x(), v.position().y(), NODE_SIZE+(agedness)*2*NODE_SIZE, NODE_SIZE+(agedness)*2*NODE_SIZE );
        }
    }
    
  }

}

void keyPressed()
{
  if ( key == 'c' )
  {
    //initialize();
    //tickDate();
    return;
  }
  
  if ( key == ' ' )
  {
    //nextNode();
    return;
  }
}

// ME ////////////////////////////////////////////

void nextNode(){
    if (fadeCounter>0){
        fadeCounter--;        

        if (fadeCounter==0 || branchNodes.size()==0){
            // clear the field
            initialize();
            branchNodes.clear();
            branchAges.clear();
            branchTags.clear();
            fadeCounter=0;
        }
    }else{
    
        try{
            JSONArray cmd = drawCode.getJSONArray(drawPtr);
            String action = cmd.getString(0);
            if (action.equals("add")){
                // find the parent's name in branch nodes
                String nodeName = cmd.getString(1);
                String parentName = cmd.getString(2);
                String tagName = cmd.getString(3);
            
                int parentIdx=-1;
                for (int i=0; i<branchNodes.size(); i++){
                    String possibleParent = (String)branchNodes.get(i);
                    if (possibleParent.equals(parentName)){
                        parentIdx = i;
                        break;
                    }
                }
                //println("Adding node "+branchNodes.size()+" to p "+parentIdx+" - "+parentName);
            
                // add the kid to branchnodes and create the new particle
                branchAges.add(new Integer(0));
                branchNodes.add(nodeName);
                branchTags.add(tagName);

                Particle newP = physics.makeParticle();
                if (parentIdx != -1){
                    Particle parentP = physics.getParticle(parentIdx);
                    addSpacersToNode(newP, parentP);
                    makeEdgeBetween(newP, parentP);
                    newP.moveTo( parentP.position().x() + random( -1, 1 ), 
                                 parentP.position().y() + random( -1, 1 ), 0 );
                }
            }else if (action.equals("date")){
                // Set the name string we should be saving image files to
                yesterdayName = theDate;
                theDate = cmd.getString(1);            
                //theDate = prettyDate(dateString);
                println("Today is "+theDate+" yesterday "+yesterdayName);
            
                fadeCounter = fadePeriod;
            }
        
            physics.tick(0.0);
            drawPtr += 1;
        
            if (drawPtr>=drawCode.length()){
                //drawPtr = drawCode.length()-1;
                drawPtr = 0;
                if (movieOn){
                    mm.finishMovie();
                    println("Done recording");
                    movieOn = false;
                }
            }
        } catch (JSONException e) { println (e.toString()); }

    }

    // Regardless of whether we're adding nodes, node should continue to age
    maxAge=0;
    for (int i=1; i<branchAges.size(); i++){
        Integer oldAge = (Integer)branchAges.get(i);
        int newAge = 1+oldAge.intValue();
        maxAge = max(maxAge, newAge);
        branchAges.set(i, new Integer(newAge));
        //println(i+": "+newAge);
    }


/*    if (fadeCounter>0){
        return;
    }

    try{
        scriptPtr += 1;
        if(scriptPtr >= drawScript.length()){
            scriptPtr = 0;
            initialize();
            branchNodes.clear();
            branchAges.clear();
        }
        JSONArray cmd = drawScript.getJSONArray(scriptPtr);
        String action = cmd.getString(0);
        if (action.equals("add")){
            // find the parent's name in branch nodes
            String nodeName = cmd.getString(1);
            String parentName = cmd.getString(2);
            String tagName = cmd.getString(3);
            
            int parentIdx=-1;
            for (int i=0; i<branchNodes.size(); i++){
                String possibleParent = (String)branchNodes.get(i);
                if (possibleParent.equals(parentName)){
                    parentIdx = i;
                    break;
                }
            }
            //println("Adding node "+branchNodes.size()+" to p "+parentIdx);
            
            // add the kid to branchnodes and create the new particle
            branchAges.add(new Integer(0));
            branchNodes.add(nodeName);
            branchTags.add(tagName);

            Particle newP = physics.makeParticle();
            if (parentIdx != -1){
                Particle parentP = physics.getParticle(parentIdx);
                addSpacersToNode(newP, parentP);
                makeEdgeBetween(newP, parentP);
                newP.moveTo( parentP.position().x() + random( -1, 1 ), 
                             parentP.position().y() + random( -1, 1 ), 0 );
            }

        }else if (action.equals("prune")){
            String pruneNode = cmd.getString(1);
            for (int i=0; i<branchNodes.size(); i++){
                String possibleCut = (String)branchNodes.get(i);
                if (possibleCut.equals(pruneNode)){
                    branchNodes.remove(i);
                    branchAges.remove(i);
                    branchTags.remove(i);
                    physics.getParticle(i).kill();
                }
                
            }
        
        }else if (action.equals("day")){
            String dateString = cmd.getString(1);
            yesterdayName = dateName;
            dateName = prettyDate(dateString);
            println("Today is "+dateName);
            theDate = dateString;
            fadeCounter=fadePeriod;
        }

        // Increment the age counts for all the nodes currently alive
        maxAge=0;
        for (int i=1; i<branchAges.size(); i++){
            Integer oldAge = (Integer)branchAges.get(i);
            int newAge = 1+oldAge.intValue();
            maxAge = max(maxAge, newAge);
            branchAges.set(i, new Integer(newAge));
            //println(i+": "+newAge);
        }

        physics.tick(0.0);
        
        
    } catch (JSONException e) {
       println (e.toString());
        //There should probably be some better exception handling here...
    }
*/    
}




color interpColor(color baseColor, color mixColor, float mixRatio){
  //println("mixing reds "+red(baseColor)+" and "+red(mixColor) + " to "+red(baseColor)*mixRatio + red(mixColor)*(1.0-mixRatio));
  return color(red(baseColor)*mixRatio + red(mixColor)*(1.0-mixRatio),
               green(baseColor)*mixRatio + green(mixColor)*(1.0-mixRatio),
               blue(baseColor)*mixRatio + blue(mixColor)*(1.0-mixRatio),
               alpha(baseColor)*mixRatio + alpha(mixColor)*(1.0-mixRatio));  
}

String prettyDate(String uglyDate){
 StringBuffer uglyBuffer = new StringBuffer(uglyDate);
 String year = uglyBuffer.substring(0,4);
 String month = uglyBuffer.substring(4,6);
 String day = uglyBuffer.substring(6,8);
 
 if(month.equals("01")){
  month = "January"; 
 }else if(month.equals("02")){
  month = "February"; 
 }else if(month.equals("03")){
  month = "March"; 
 }else if(month.equals("04")){
  month = "April"; 
 }else if(month.equals("05")){
  month = "May"; 
 }else if(month.equals("06")){
  month = "June"; 
 }else if(month.equals("07")){
  month = "July"; 
 }else if(month.equals("08")){
  month = "August"; 
 }else if(month.equals("09")){
  month = "September"; 
 }else if(month.equals("10")){
  month = "October"; 
 }else if(month.equals("11")){
  month = "November"; 
 }else if(month.equals("12")){
  month = "December"; 
 }
 
 if(uglyBuffer.substring(6,7).equals("0")){
   day = uglyBuffer.substring(7,8);   
 }
 
 //return day+" "+month+" "+year; 
 return month+" "+year; 
}

void updateCentroid()
{
  float minWidth = 50;
  float minHeight = 50;
  float 
    xMax = Float.NEGATIVE_INFINITY, 
    xMin = Float.POSITIVE_INFINITY, 
    yMin = Float.POSITIVE_INFINITY, 
    yMax = Float.NEGATIVE_INFINITY;

  for ( int i = 0; i < physics.numberOfParticles(); ++i )
  {
    Particle p = physics.getParticle( i );
    xMax = max( xMax, p.position().x() );
    xMin = min( xMin, p.position().x() );
    yMin = min( yMin, p.position().y() );
    yMax = max( yMax, p.position().y() );
  }
  float deltaX = xMax-xMin;
  float deltaY = yMax-yMin;
  
  //println("particle xrange: "+xMin+" - "+xMax);
  //println("         yrange: "+yMin+" - "+yMax);
  //println("world size: "+deltaX+" x "+deltaY);

  // let's zoom out a little bit by exaggerating the dimensions slightly
  int increment = (int)(max(deltaX,deltaY)*0.55);
  increment = (int)max(increment, 280.0);
  
  
  xMin -= increment;
  yMin -= increment;
  deltaX += 2*increment;
  deltaY += 2*increment;
  
  if ( deltaY > deltaX )
    centroid.setTarget( xMin-10 + 0.5*deltaX, yMin +0.5*deltaY, height/(deltaY+50) );
  else
    centroid.setTarget( xMin-10 + 0.5*deltaX, yMin +0.5*deltaY, width/(deltaX+50) );
}

void addSpacersToNode( Particle p, Particle r )
{
  for ( int i = 0; i < physics.numberOfParticles(); ++i )
  {
    Particle q = physics.getParticle( i );
    if ( p != q && p != r )
      physics.makeAttraction( p, q, -SPACER_STRENGTH, 20 );
  }
}

void makeEdgeBetween( Particle a, Particle b )
{
  physics.makeSpring( a, b, EDGE_STRENGTH, EDGE_STRENGTH, EDGE_LENGTH );
}

void initialize()
{
  physics.clear();
  //physics.makeParticle();
  centroid.setValue( 0, 0, 1.0 );
}


